@@ -1,6 +1,6 @@
 /*
     MIDI Virtual Piano Keyboard
-    Copyright (C) 2008-2020, Pedro Lopez-Cabanillas <plcl@users.sf.net>
+    Copyright (C) 2008-2021, Pedro Lopez-Cabanillas <plcl@users.sf.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,8 +16,10 @@
     with this program; If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <QDebug>
 #include <QDesktopServices>
 #include <QInputDialog>
+#include <QFileDialog>
 #include <QDir>
 #include <QMessageBox>
 #include <QApplication>
@@ -37,7 +39,8 @@
 #include <QTranslator>
 #include <QMapIterator>
 #include <QShortcut>
-#include <QDebug>
+#include <QActionGroup>
+#include <QStandardPaths>
 
 #include <drumstick/backendmanager.h>
 #include <drumstick/rtmidiinput.h>
@@ -87,21 +90,23 @@ VPiano::VPiano( QWidget * parent, Qt::WindowFlags flags )
     m_trq = new QTranslator(this);
     m_trp = new QTranslator(this);
     m_trl = new QTranslator(this);
-    if (!m_trq->load(QSTR_QTPX + VPianoSettings::instance()->language(), VPianoSettings::systemLocales())) {
-        qWarning() << "Failure loading Qt5 system translations for" << VPianoSettings::instance()->language()
+    QString lang = VPianoSettings::instance()->language();
+    if (!m_trq->load(QSTR_QTPX + lang, VPianoSettings::systemLocales()) && !lang.startsWith("en")) {
+        qWarning() << "Failure loading Qt5 system translations for" << lang
                    << "from" << VPianoSettings::systemLocales();
     }
-    if (!m_trp->load(QSTR_VMPKPX + VPianoSettings::instance()->language(), VPianoSettings::localeDirectory())) {
-        qWarning() << "Failure loading VMPK application translations for" << VPianoSettings::instance()->language()
+    if (!m_trp->load(QSTR_VMPKPX + lang, VPianoSettings::localeDirectory()) && !lang.startsWith("en")) {
+        qWarning() << "Failure loading VMPK application translations for" << lang
                    << "from" << VPianoSettings::localeDirectory();
     }
-    if (!m_trl->load(QSTR_DRUMSTICKPX + VPianoSettings::instance()->language(), VPianoSettings::drumstickLocales())) {
-        qWarning() << "Failure loading widgets library translations for" << VPianoSettings::instance()->language()
+    if (!m_trl->load(QSTR_DRUMSTICKPX + lang, VPianoSettings::drumstickLocales()) && !lang.startsWith("en")) {
+        qWarning() << "Failure loading widgets library translations for" << lang
                    << "from" << VPianoSettings::drumstickLocales();
     }
     QCoreApplication::installTranslator(m_trq);
     QCoreApplication::installTranslator(m_trp);
     QCoreApplication::installTranslator(m_trl);
+    VPianoSettings::instance()->retranslatePalettes();
     ui.setupUi(this);
     initLanguages();
 
@@ -145,6 +150,8 @@ VPiano::VPiano( QWidget * parent, Qt::WindowFlags flags )
     connect(ui.actionColorPalette, SIGNAL(triggered()), SLOT(slotColorPolicy()));
     connect(ui.actionColorScale, SIGNAL(toggled(bool)), SLOT(slotColorScale(bool)));
     connect(ui.actionWindowFrame, SIGNAL(toggled(bool)), SLOT(toggleWindowFrame(bool)));
+    connect(ui.actionLoad_Configuration, &QAction::triggered, this, &VPiano::slotLoadConfiguration);
+    connect(ui.actionSave_Configuration, &QAction::triggered, this, &VPiano::slotSaveConfiguration);
     // Toolbars actions: toggle view
     connect(ui.toolBarNotes->toggleViewAction(), SIGNAL(toggled(bool)),
             ui.actionNotes, SLOT(setChecked(bool)));
@@ -212,53 +219,76 @@ bool VPiano::initMidi()
     m_backendManager = new BackendManager();
     m_backendManager->refresh(VPianoSettings::instance()->settingsMap());
 
-    m_midiin = m_backendManager->inputBackendByName(VPianoSettings::instance()->lastInputBackend());
+    m_midiin = m_backendManager->findInput(VPianoSettings::instance()->lastInputBackend());
     if (m_midiin == nullptr) {
-        if (VPianoSettings::instance()->lastInputBackend() != VPianoSettings::instance()->nativeInput()) {
-            qWarning() << "Last MIDI IN driver" << VPianoSettings::instance()->lastInputBackend() << "not found";
-        }
-        m_midiin = m_backendManager->inputBackendByName(VPianoSettings::instance()->nativeInput());
-        if (m_midiin == nullptr) {
-            qWarning() << "Default MIDI IN driver" << VPianoSettings::instance()->nativeInput() << "not found";
-        }
+        qWarning() << "MIDI IN driver not available";
     }
 
-    m_midiout = m_backendManager->outputBackendByName(VPianoSettings::instance()->lastOutputBackend());
+    m_midiout = m_backendManager->findOutput(VPianoSettings::instance()->lastOutputBackend());
     if (m_midiout == nullptr) {
-        if (VPianoSettings::instance()->lastOutputBackend() != VPianoSettings::instance()->nativeOutput()) {
-            qWarning() << "Last MIDI OUT driver" << VPianoSettings::instance()->lastOutputBackend() << "not found";
-        }
-        m_midiout = m_backendManager->outputBackendByName(VPianoSettings::instance()->nativeOutput());
-        if (m_midiout == nullptr) {
-            qWarning() << "Default MIDI OUT driver" << VPianoSettings::instance()->nativeOutput() << "not found";
-        }
+        qWarning() << "MIDI OUT driver not available";
     }
 
     SettingsFactory settings;
 
     if (m_midiin != nullptr) {
-        if (!VPianoSettings::instance()->lastInputConnection().isEmpty()) {
-            m_midiin->initialize(settings.getQSettings());
-            for(const MIDIConnection& conn : m_midiin->connections()) {
-                if (conn.first == VPianoSettings::instance()->lastInputConnection()) {
-                    m_midiin->open(conn);
-                    break;
+        connectMidiInSignals();
+        m_midiin->initialize(settings.getQSettings());
+        MIDIConnection conn;
+        auto connections = m_midiin->connections();
+        auto lastConn = VPianoSettings::instance()->lastInputConnection();
+        auto itr = std::find_if(connections.constBegin(), connections.constEnd(), [lastConn](const MIDIConnection& c){return c.first == lastConn;});
+        if (itr == connections.constEnd()) {
+            if (!connections.isEmpty()) {
+                conn = connections.first();
+            }
+        } else {
+            conn = (*itr);
+        }
+        m_midiin->open(conn);
+        auto metaObj = m_midiin->metaObject();
+        if ((metaObj->indexOfProperty("status") != -1) &&
+            (metaObj->indexOfProperty("diagnostics") != -1)) {
+            auto status = m_midiin->property("status");
+            if (status.isValid() && !status.toBool()) {
+                auto diagnostics = m_midiin->property("diagnostics");
+                if (diagnostics.isValid()) {
+                    auto text = diagnostics.toStringList().join(QChar::LineFeed).trimmed();
+                    qWarning() << "MIDI Input" << text;
                 }
             }
         }
     }
 
-    if (m_midiout != nullptr && !VPianoSettings::instance()->lastOutputConnection().isEmpty()) {
+    if (m_midiout != nullptr) {
         m_midiout->initialize(settings.getQSettings());
-        for(const MIDIConnection& conn : m_midiout->connections()) {
-            if (conn.first == VPianoSettings::instance()->lastOutputConnection()) {
-                m_midiout->open(conn);
-                if (!VPianoSettings::instance()->lastInputConnection().isEmpty()) {
-                    m_midiin->setMIDIThruDevice(m_midiout);
-                    m_midiin->enableMIDIThru(VPianoSettings::instance()->midiThru());
-                }
-                break;
+        MIDIConnection conn;
+        auto connections = m_midiout->connections();
+        auto lastConn = VPianoSettings::instance()->lastOutputConnection();
+        auto itr = std::find_if(connections.constBegin(), connections.constEnd(), [lastConn](const MIDIConnection& c){return c.first == lastConn;});
+        if (itr == connections.constEnd()) {
+            if (!connections.isEmpty()) {
+                conn = connections.first();
             }
+        } else {
+            conn = (*itr);
+        }
+        m_midiout->open(conn);
+        auto metaObj = m_midiout->metaObject();
+        if ((metaObj->indexOfProperty("status") != -1) &&
+            (metaObj->indexOfProperty("diagnostics") != -1)) {
+            auto status = m_midiout->property("status");
+            if (status.isValid() && !status.toBool()) {
+                auto diagnostics = m_midiout->property("diagnostics");
+                if (diagnostics.isValid()) {
+                    auto text = diagnostics.toStringList().join(QChar::LineFeed).trimmed();
+                    qWarning() << "MIDI Output" << text;
+                }
+            }
+        }
+        if (m_midiin != nullptr) {
+            m_midiin->setMIDIThruDevice(m_midiout);
+            m_midiin->enableMIDIThru(VPianoSettings::instance()->midiThru());
         }
     }
 
@@ -337,7 +367,8 @@ void VPiano::initToolBars()
     ui.toolBarControllers->addWidget(m_lblControl);
     m_lblControl ->setMargin(TOOLBARLABELMARGIN);
     m_comboControl = new QComboBox(this);
-    m_comboControl->setStyleSheet("combobox-popup: 0;");
+    m_comboControl->setObjectName("cboControl");
+    //m_comboControl->setStyleSheet("combobox-popup: 0;");
     m_comboControl->setSizeAdjustPolicy(QComboBox::AdjustToContents);
     m_comboControl->setFocusPolicy(Qt::NoFocus);
     ui.toolBarControllers->addWidget(m_comboControl);
@@ -378,14 +409,16 @@ void VPiano::initToolBars()
     ui.toolBarPrograms->addWidget(m_lblBank);
     m_lblBank->setMargin(TOOLBARLABELMARGIN);
     m_comboBank = new QComboBox(this);
+    m_comboBank->setObjectName("cboBank");
     m_comboBank->setSizeAdjustPolicy(QComboBox::AdjustToContents);
-    m_comboBank->setStyleSheet("combobox-popup: 0;");
+    //m_comboBank->setStyleSheet("combobox-popup: 0;");
     m_comboBank->setFocusPolicy(Qt::NoFocus);
     ui.toolBarPrograms->addWidget(m_comboBank);
     m_lblProgram = new QLabel(this);
     ui.toolBarPrograms->addWidget(m_lblProgram);
     m_lblProgram->setMargin(TOOLBARLABELMARGIN);
     m_comboProg = new QComboBox(this);
+    m_comboProg->setObjectName("cboProg");
     m_comboProg->setStyleSheet("combobox-popup: 0;");
     m_comboProg->setSizeAdjustPolicy(QComboBox::AdjustToContents);
     m_comboProg->setFocusPolicy(Qt::NoFocus);
@@ -569,8 +602,6 @@ void VPiano::initExtraControllers()
 void VPiano::readSettings()
 {
     VPianoSettings::instance()->retranslatePalettes();
-    restoreGeometry(VPianoSettings::instance()->geometry());
-    restoreState(VPianoSettings::instance()->state());
 
     ui.actionStatusBar->setChecked(VPianoSettings::instance()->showStatusBar());
     ui.statusBar->setVisible(VPianoSettings::instance()->showStatusBar());
@@ -766,7 +797,7 @@ void VPiano::writeSettings()
 
 void VPiano::closeEvent( QCloseEvent *event )
 {
-    //qDebug() << "closeEvent:" << event->type();
+    //qDebug() << Q_FUNC_INFO;
     if (m_initialized) {
         writeSettings();
     }
@@ -797,6 +828,8 @@ QColor VPiano::getHighlightColorFromPolicy(const int chan, const int note, const
         return palette.getColor(getType(note));
     case PAL_CHANNELS:
         return palette.getColor(chan);
+    case PAL_HISCALE:
+        return palette.getColor(getDegree(note));
     default:
         break;
     }
@@ -903,7 +936,17 @@ void VPiano::slotPitchBend(const int chan, const int value)
 
 void VPiano::showEvent ( QShowEvent *event )
 {
-    //qDebug() << "showEvent:" << event->type();
+    static bool firstTime{true};
+    //qDebug() << Q_FUNC_INFO << firstTime;
+    if (firstTime) {
+        if (!restoreGeometry(VPianoSettings::instance()->geometry())) {
+            qWarning() << "restoreGeometry() failed!";
+        }
+        if (!restoreState(VPianoSettings::instance()->state())) {
+            qWarning() << "restoreState() failed!";
+        }
+        firstTime = false;
+    }
     QMainWindow::showEvent(event);
 #if !defined(SMALL_SCREEN)
     if (m_initialized) {
@@ -1153,6 +1196,19 @@ void VPiano::slotAboutQt()
     grabKb();
 }
 
+void VPiano::connectMidiInSignals()
+{
+    if (m_midiin != nullptr) {
+        connect(m_midiin, &MIDIInput::midiNoteOn, this, &VPiano::slotNoteOn, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiNoteOff, this, &VPiano::slotNoteOff, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiKeyPressure, this, &VPiano::slotKeyPressure, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiChannelPressure, this, &VPiano::slotChannelPressure, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiController, this, &VPiano::slotController, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiProgram, this, &VPiano::slotProgram, Qt::QueuedConnection);
+        connect(m_midiin, &MIDIInput::midiPitchBend, this, &VPiano::slotPitchBend, Qt::QueuedConnection);
+    }
+}
+
 void VPiano::slotConnections()
 {
     QPointer<MidiSetup> dlgMidiSetup = new MidiSetup(this);
@@ -1169,19 +1225,9 @@ void VPiano::slotConnections()
         if (m_midiout != nullptr) {
             m_midiout->disconnect();
         }
-
         m_midiin = dlgMidiSetup->getInput();
         m_midiout = dlgMidiSetup->getOutput();
-
-        if (m_midiin != nullptr) {
-            connect(m_midiin, SIGNAL(midiNoteOn(int,int,int)), SLOT(slotNoteOn(int,int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiNoteOff(int,int,int)), SLOT(slotNoteOff(int,int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiKeyPressure(int,int,int)), SLOT(slotKeyPressure(int,int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiChannelPressure(int,int)), SLOT(slotChannelPressure(int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiController(int,int,int)), SLOT(slotController(int,int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiProgram(int,int)), SLOT(slotProgram(int,int)), Qt::QueuedConnection);
-            connect(m_midiin, SIGNAL(midiPitchBend(int,int)), SLOT(slotPitchBend(int,int)), Qt::QueuedConnection);
-        }
+        connectMidiInSignals();
         enforceMIDIChannelState();
     }
     grabKb();
@@ -1232,6 +1278,9 @@ void VPiano::populateControllers()
 
 void VPiano::applyPreferences()
 {
+    static QPalette defaultPalette = qApp->palette();
+    static QPalette darkPalette(QColor(0x30,0x30,0x30));
+
     ui.pianokeybd->allKeysOff();
     ui.pianokeybd->setFont(VPianoSettings::instance()->namesFont());
     ui.pianokeybd->setLabelOctave(VPianoSettings::instance()->namesOctave());
@@ -1256,6 +1305,12 @@ void VPiano::applyPreferences()
     ui.actionMouseInput->setChecked(enableMouse);
     ui.actionTouchScreenInput->setChecked(enableTouch);
 
+#if defined(Q_OS_WINDOWS)
+    m_snapper.SetEnabled(VPianoSettings::instance()->getWinSnap());
+#endif
+    qApp->setPalette( VPianoSettings::instance()->getDarkMode() ? darkPalette : defaultPalette );
+    qApp->setStyle( VPianoSettings::instance()->getStyle() );
+
     VMPKKeyboardMap* map = VPianoSettings::instance()->getKeyboardMap();
     if (!map->getFileName().isEmpty() && map->getFileName() != QSTR_DEFAULT )
         ui.pianokeybd->setKeyboardMap(map);
@@ -1896,10 +1951,11 @@ void VPiano::slotControllerDown()
 void VPiano::slotSwitchLanguage(QAction *action)
 {
     QString lang = action->data().toString();
-    if ( QMessageBox::question (this, tr("Language Changed"),
+    bool result = QMessageBox::question (this, tr("Language Changed"),
             tr("The language for this application is going to change to %1. "
                "Do you want to continue?").arg(m_supportedLangs[lang]),
-            QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes )
+            QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes;
+    if ( result )
     {
         VPianoSettings::instance()->setLanguage(lang);
         retranslateUi();
@@ -1913,8 +1969,7 @@ void VPiano::createLanguageMenu()
     QString currentLang = VPianoSettings::instance()->language();
     QActionGroup *languageGroup = new QActionGroup(this);
     languageGroup->setExclusive(true);
-    connect(languageGroup, SIGNAL(triggered(QAction *)),
-            SLOT(slotSwitchLanguage(QAction *)));
+    connect(languageGroup, &QActionGroup::triggered, this, &VPiano::slotSwitchLanguage, Qt::QueuedConnection);
     QDir dir(VPianoSettings::localeDirectory());
     QStringList fileNames = dir.entryList(QStringList(QSTR_VMPKPX + "*.qm"));
     QStringList locales;
@@ -2169,3 +2224,69 @@ void VPiano::slotNoteName(const QString& name)
         ui.statusBar->showMessage(name);
     }
 }
+
+void VPiano::slotLoadConfiguration()
+{
+    releaseKb();
+    auto configDir = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation);
+    QString fileName = QFileDialog::getOpenFileName(this, tr("Open Configuration File"),
+        configDir, tr("Configuration files (*.conf *.ini)"));
+    if (!fileName.isEmpty()) {
+        VPianoSettings::instance()->ReadFromFile(fileName);
+        qApp->exit(RESTART_VMPK);
+    }
+    grabKb();
+}
+
+void VPiano::slotSaveConfiguration()
+{
+    releaseKb();
+    auto configDir = QStandardPaths::writableLocation(QStandardPaths::AppConfigLocation);
+    QFileDialog dlg(this);
+    dlg.setNameFilter(tr("Configuration files (*.conf *.ini)"));
+    dlg.setDirectory(configDir);
+    dlg.setWindowTitle(tr("Save Configuration File"));
+    dlg.setDefaultSuffix("conf");
+    dlg.setFileMode(QFileDialog::AnyFile);
+    dlg.setAcceptMode(QFileDialog::AcceptSave);
+    if (dlg.exec() == QDialog::Accepted) {
+        auto selected = dlg.selectedFiles();
+        if (!selected.isEmpty()) {
+            VPianoSettings::instance()->SaveToFile(selected.first());
+        }
+    }
+    grabKb();
+}
+
+#if QT_VERSION < QT_VERSION_CHECK(6,0,0)
+bool VPiano::nativeEvent(const QByteArray &eventType, void *message, long *result)
+#else
+bool VPiano::nativeEvent(const QByteArray &eventType, void *message, qintptr *result)
+#endif
+{
+#if defined(Q_OS_WINDOWS)
+    if (VPianoSettings::instance()->getWinSnap() && m_snapper.HandleMessage(message)) {
+        result = 0;
+        return true;
+    }
+#endif
+    return QWidget::nativeEvent(eventType, message, result);
+}
+
+void VPiano::changeEvent(QEvent *event)
+{
+#if defined(Q_OS_WINDOWS)
+    if (event->type() == QEvent::PaletteChange) {
+        foreach(QToolBar *tb, findChildren<QToolBar*>()) {
+            foreach(QComboBox *cb, tb->findChildren<QComboBox*>()) {
+                cb->setPalette(qApp->palette());
+                foreach(QWidget *w, cb->findChildren<QWidget*>()) {
+                    w->setPalette(qApp->palette());
+                }
+                //qDebug() << cb;
+            }
+        }
+    }
+#endif
+    QMainWindow::changeEvent(event);
+}
